# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property
from typing import Any, Protocol, Self, TypedDict


class RunnerLoggingExtras(TypedDict):
    id: str


class Event(Protocol):
    def is_set(self) -> bool: ...  # pragma: nocover

    def set(self) -> None: ...  # pragma: nocover

    def wait(self, timeout: float | None = None) -> bool: ...  # pragma: nocover


class Runner(Protocol):
    @cached_property
    def name(self) -> str: ...  # pragma: nocover

    @cached_property
    def logging_extras(self) -> RunnerLoggingExtras: ...  # pragma: nocover

    def wait(self) -> None: ...  # pragma: nocover

    def signal(self, sig: int | None = None) -> None: ...  # pragma: nocover

    def __enter__(self) -> Self: ...  # pragma: nocover

    def __exit__(self, *_: Any, **__: Any) -> None: ...  # pragma: nocover



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import threading
from dataclasses import dataclass
from typing import Any

from quesadilla.connectors.basic_queues import BasicQueue
from quesadilla.connectors.in_memory.thread_safe import Queue
from quesadilla.core.task_metadata import SerializableTask

from quesadilla.runners.runners.base import BaseRunner
from quesadilla.runners.runners.protocol import Event
from quesadilla.runners.workloads.protocol import Executor, Workload


class Thread[T: Workload](threading.Thread, Executor[T]):
    def __init__(self, workload: T, *args: Any, **kwargs: Any) -> None:
        super().__init__(name=workload.name, *args, **kwargs)
        self.workload = workload

    def run(self) -> None:
        self.workload.main_loop()

    def close(self) -> None:
        self.workload.close()

    def stop(self) -> None:
        self.workload.stop()


@dataclass(frozen=True)
class MultithreadingRunner(BaseRunner):
    def make_workload[T: Workload](self, workload: T) -> Executor[T]:
        return Thread(workload)

    def make_queue(self) -> BasicQueue[SerializableTask[Any]]:
        return Queue()

    def make_event(self) -> Event:
        return threading.Event()



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import logging
import multiprocessing
import signal
import threading
from types import FrameType
from typing import Annotated, TypedDict

import typer
from quesadilla.cli.utils import task_queue
from quesadilla.core.logging import TaskQueueLoggingExtras
from quesadilla.core.task_queues import TaskQueue

from quesadilla.runners import (
    MultiprocessingRunner,
    MultithreadingRunner,
    RandomSelector,
    RoundRobinSelector,
    RunnerConfig,
)


class Parallelism(enum.StrEnum):
    THREADS = enum.auto()
    PROCESSES = enum.auto()


class WorkerSelector(enum.StrEnum):
    ROUNDROBIN = enum.auto()
    RANDOM = enum.auto()


class ListenerLoggingExtras(TypedDict):
    task_queue: TaskQueueLoggingExtras
    parallelism: str
    brokers: int
    workers: int
    worker_selector: str


def listener(
    task_queue: Annotated[
        TaskQueue,
        typer.Argument(
            parser=task_queue,
            help=("Dotted path of the TaskQueue object <path.to.module::task_queue>"),
            show_default=False,
        ),
    ],
    parallelism: Annotated[
        Parallelism, typer.Option(help="Parallelization type")
    ] = Parallelism.THREADS,
    brokers: Annotated[int, typer.Option(help="Number of broker threads")] = 1,
    workers: Annotated[int, typer.Option(help="Number of worker threads")] = 1,
    worker_selector: Annotated[
        WorkerSelector, typer.Option(help="Worker selector implementation")
    ] = WorkerSelector.ROUNDROBIN,
) -> None:
    """
    Run a listener process for processing tasks
    """
    match worker_selector:
        case WorkerSelector.ROUNDROBIN:
            worker_selector_class = RoundRobinSelector
        case WorkerSelector.RANDOM:
            worker_selector_class = RandomSelector

    runner_config = RunnerConfig(
        brokers=brokers,
        workers=workers,
        worker_selector_class=worker_selector_class,
    )

    match parallelism:
        case Parallelism.THREADS:
            stopped = threading.Event()
            runner = MultithreadingRunner(task_queue, config=runner_config)
        case Parallelism.PROCESSES:
            stopped = multiprocessing.get_context("spawn").Event()
            runner = MultiprocessingRunner(task_queue, config=runner_config)

    logger = logging.getLogger(__name__)
    logging_extras: ListenerLoggingExtras = {
        "task_queue": task_queue.logging_extras,
        "parallelism": parallelism.value,
        "brokers": brokers,
        "workers": workers,
        "worker_selector": worker_selector.value,
    }

    def shutdown(sig: int | None = None, frame: FrameType | None = None) -> None:
        if not stopped.is_set():
            s = signal.Signals(sig) if sig is not None else sig
            if s is not None:
                logger.warning(f"Received signal {s.name}", extra=logging_extras)
            if s in frozenset((signal.SIGTERM, signal.SIGINT)):
                logger.info("Listener stopping...", extra=logging_extras)
                stopped.set()
                runner.signal(sig)

    logger.info("Listener starting...", extra=logging_extras)

    with runner:
        signal.signal(signal.SIGINT, shutdown)
        signal.signal(signal.SIGTERM, shutdown)
        runner.wait()

    logger.info("Listener exited gracefully", extra=logging_extras)



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random
from collections.abc import Iterable, Iterator
from dataclasses import InitVar, dataclass, field
from typing import Protocol, Self

from quesadilla.runners.workloads.workers import Worker


class WorkerSelector(Protocol):
    def __init__(self, collection: list[Worker]) -> None: ...  # pragma: nocover

    def select(self) -> Worker: ...  # pragma: nocover

    def all(self) -> Iterable[Worker]: ...  # pragma: nocover


@dataclass(frozen=True)
class RandomSelector(WorkerSelector):
    collection: list[Worker] = field(repr=False)

    def select(self) -> Worker:
        return random.choice(self.collection)  # nosec

    def all(self) -> list[Worker]:
        return self.collection


@dataclass(frozen=True)
class RoundRobinSelector(WorkerSelector):
    @dataclass(frozen=True)
    class CircularIterator[IT]:
        collection: Iterable[IT] = field(repr=False)
        current_iterator: Iterator[IT] = field(init=False, repr=False)

        def __post_init__(self) -> None:
            self.reset()

        def __iter__(self) -> Self:
            return self

        def __next__(self) -> IT:
            try:
                return next(self.current_iterator)
            except StopIteration:
                self.reset()
                return next(self.current_iterator)

        def reset(self) -> None:
            super().__setattr__("current_iterator", iter(self.collection))

    collection: InitVar[list[Worker]]
    iterator: CircularIterator[Worker] = field(init=False, repr=False)

    def __post_init__(self, collection: list[Worker]) -> None:
        super().__setattr__(
            "iterator",
            iter(RoundRobinSelector.CircularIterator(collection)),
        )

    def select(self) -> Worker:
        return next(self.iterator)

    def all(self) -> Iterable[Worker]:
        return self.iterator.collection



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import multiprocessing
import multiprocessing.context
import multiprocessing.managers
import multiprocessing.queues
import multiprocessing.synchronize
import signal
import time
from dataclasses import dataclass, field
from typing import Any

from quesadilla.connectors.basic_queues import BasicQueue
from quesadilla.connectors.in_memory.process_safe import Queue
from quesadilla.core.task_metadata import SerializableTask

from quesadilla.runners.runners.base import BaseRunner
from quesadilla.runners.runners.protocol import Event
from quesadilla.runners.workloads.protocol import Executor, Workload


class Process[T: Workload](multiprocessing.Process, Executor[T]):
    def __init__(self, workload: T, *args: Any, **kwargs: Any) -> None:
        super().__init__(name=workload.name, *args, **kwargs)
        self.workload = workload

    def run(self) -> None:  # pragma: nocover
        # this function is excluded from coverage
        # because it's executed in a separate process
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)

        self.workload.main_loop()

    def close(self) -> None:
        self.workload.close()

    def stop(self) -> None:
        self.workload.stop()


@dataclass(frozen=True)
class MultiprocessingRunner(BaseRunner):
    object_manager: multiprocessing.managers.SyncManager = field(
        default_factory=multiprocessing.get_context("spawn").Manager,
        init=False,
        repr=False,
    )

    def wait(self) -> None:
        # @hack: this is a workaround for a deadlock with multiprocessing.Event.wait
        while not self.stopped.wait(timeout=0.1):  # pragma: nobranch
            time.sleep(0.5)  # pragma: nocover

    def make_workload[T: Workload](self, workload: T) -> Executor[T]:
        return Process(workload)

    def make_queue(self) -> BasicQueue[SerializableTask[Any]]:
        return Queue(self.object_manager.Queue())  # type: ignore

    def make_event(self) -> Event:
        return self.object_manager.Event()



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from quesadilla.runners.runners.composite import CompositeRunner
from quesadilla.runners.runners.configs import RunnerConfig
from quesadilla.runners.runners.multiprocessing import MultiprocessingRunner
from quesadilla.runners.runners.multithreading import MultithreadingRunner
from quesadilla.runners.runners.protocol import Runner

__all__ = [
    "CompositeRunner",
    "MultiprocessingRunner",
    "MultithreadingRunner",
    "Runner",
    "RunnerConfig",
]



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import time
from dataclasses import dataclass, field
from functools import cached_property

from quesadilla.core.task_queues import GenericQueuedTask

from quesadilla.runners.workloads.base import BaseWorkload
from quesadilla.runners.workloads.worker_selectors import WorkerSelector


@dataclass(frozen=True)
class Broker(BaseWorkload):
    worker_selector: WorkerSelector = field(kw_only=True)

    @cached_property
    def name(self) -> str:
        return f"{self.runner.name}.broker.{str(self.id)}"

    def main_loop(self) -> None:
        self.logger.info("Broker starting...", extra=self.logging_extras)

        while not self.stopped.is_set():
            # handling results is more important than scheduling new work
            self.handle_outstanding_tasks()

            if not self.closed.is_set() and (queued_task := self.next()) is not None:
                self.schedule_for_execution(queued_task)
                continue

            time.sleep(0.1)

        self.logger.info(
            "Broker handling outstanding tasks...", extra=self.logging_extras
        )
        self.handle_outstanding_tasks()

        self.logger.info("Broker exited gracefully", extra=self.logging_extras)

    def next(self) -> GenericQueuedTask | None:
        return self.task_queue.pull()

    def schedule_for_execution(self, queued_task: GenericQueuedTask) -> None:
        queued_task = queued_task.start()
        self.task_queue.update(queued_task)
        self.worker_selector.select().schedule_for_execution(queued_task)
        self.logger.info(
            "Scheduled task %s(%s)",
            queued_task.task_executable.name,
            queued_task.id,
            extra=self.get_queued_task_logging_extras(queued_task),
        )

    def handle_outstanding_tasks(self) -> None:
        for worker in self.worker_selector.all():
            while task := worker.pull_result():
                self.result(task)

    def result(self, queued_task: GenericQueuedTask) -> None:
        if queued_task.execution_result.finished:
            queued_task = queued_task.finalize()
        else:
            queued_task = queued_task.reset()

        self.task_queue.update(queued_task)
        self.logger.info(
            "Finalized task %s(%s)",
            queued_task.task_executable.name,
            queued_task.id,
            extra=self.get_queued_task_logging_extras(queued_task),
        )

    def close(self):
        self.logger.info("Broker closing...", extra=self.logging_extras)
        super().close()

    def stop(self):
        self.logger.info("Broker stopping...", extra=self.logging_extras)
        super().stop()



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from dataclasses import dataclass, field
from functools import cached_property

import ulid
from quesadilla.core.task_queues import GenericQueuedTask, TaskQueue

from quesadilla.runners.runners.protocol import Event, Runner
from quesadilla.runners.workloads.protocol import (
    ExecutedQueuedTaskLoggingExtras,
    Workload,
    WorkloadLoggingExtras,
)


@dataclass(frozen=True)
class BaseWorkload(Workload):
    runner: Runner
    task_queue: TaskQueue

    id: ulid.ULID = field(default_factory=ulid.ULID, init=False)

    closed: Event = field(kw_only=True, repr=False)
    stopped: Event = field(kw_only=True, repr=False)

    logger: logging.Logger = field(init=False, repr=False)

    @cached_property
    def logging_extras(self) -> WorkloadLoggingExtras:
        return {
            "task_queue": self.task_queue.logging_extras,
            "id": str(self.id),
            "runner": self.runner.logging_extras,
        }

    def __post_init__(self) -> None:
        super().__setattr__("logger", logging.getLogger(self.name))

    def get_queued_task_logging_extras(
        self, queued_task: GenericQueuedTask
    ) -> ExecutedQueuedTaskLoggingExtras:
        return {
            **queued_task.logging_extras,
            "workload": self.logging_extras,
        }

    def close(self):
        self.closed.set()

    def stop(self):
        self.stopped.set()



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import time
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any

from quesadilla.connectors.basic_queues import BaseQueue, BasicQueue
from quesadilla.core.task_executables import Failure, Success
from quesadilla.core.task_metadata import SerializableTask
from quesadilla.core.task_queues import GenericQueuedTask

from quesadilla.runners.workloads.base import BaseWorkload


@dataclass(frozen=True)
class Worker(BaseWorkload):
    scheduled_tasks: BasicQueue[SerializableTask[Any]] = field(kw_only=True, repr=False)
    results: BasicQueue[SerializableTask[Any]] = field(kw_only=True, repr=False)

    asyncio_runner: asyncio.Runner = field(default_factory=asyncio.Runner, init=False)

    @cached_property
    def name(self) -> str:
        return f"{self.runner.name}.worker.{str(self.id)}"

    def main_loop(self) -> None:
        self.logger.info("Worker starting...", extra=self.logging_extras)

        while not self.stopped.is_set():
            while not self.closed.is_set() and (queued_task := self.next()) is not None:
                self.logger.info(
                    "Processing task %s(%s)",
                    queued_task.task_executable.name,
                    queued_task.id,
                    extra=self.get_queued_task_logging_extras(queued_task),
                )
                queued_task = queued_task.execute(self.asyncio_runner)
                self.result(queued_task)
                match queued_task.result:
                    case Success():
                        self.logger.info(
                            "Processed task %s(%s)",
                            queued_task.task_executable.name,
                            queued_task.id,
                            extra=self.get_queued_task_logging_extras(queued_task),
                        )
                    case Failure(exc=exc):  # pragma: nobranch
                        self.logger.exception(
                            "Processed task %s(%s) with an exception: %s",
                            queued_task.task_executable.name,
                            queued_task.id,
                            exc,
                            extra=self.get_queued_task_logging_extras(queued_task),
                        )

            time.sleep(0.1)

        self.logger.info("Worker flushing tasks...", extra=self.logging_extras)
        self.flush_outstanding_tasks()

        self.logger.info("Worker exited gracefully", extra=self.logging_extras)

    def flush_outstanding_tasks(self) -> SerializableTask[Any] | None:
        while (queued_task := self.next()) is not None:
            self.results.put(queued_task.serializable)

    def next(self) -> GenericQueuedTask | None:
        try:
            return GenericQueuedTask.from_serializable(
                (task := self.scheduled_tasks.get(timeout=0.1)),
                self.task_queue,
                self.task_queue.get_task_executable(task.task_name),
            )
        except BaseQueue.Empty:
            return None

    def result(self, queued_task: GenericQueuedTask) -> None:
        self.results.put(queued_task.serializable)

    def schedule_for_execution(self, task: GenericQueuedTask) -> None:
        self.scheduled_tasks.put(task.serializable)

    def pull_result(self) -> GenericQueuedTask | None:
        try:
            return GenericQueuedTask.from_serializable(
                (task := self.results.get(timeout=0.1)),
                self.task_queue,
                self.task_queue.get_task_executable(task.task_name),
            )
        except BaseQueue.Empty:
            return None

    def close(self):
        self.logger.info("Worker closing...", extra=self.logging_extras)
        super().close()

    def stop(self):
        self.logger.info("Worker stopping...", extra=self.logging_extras)
        super().stop()



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import signal
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from functools import cached_property
from itertools import chain
from typing import Any, Self

import ulid
from quesadilla.connectors.basic_queues import BasicQueue
from quesadilla.core.logging import TaskQueueLoggingExtras
from quesadilla.core.task_metadata import SerializableTask
from quesadilla.core.task_queues import TaskQueue

from quesadilla.runners.runners.configs import RunnerConfig
from quesadilla.runners.runners.protocol import Event, Runner, RunnerLoggingExtras
from quesadilla.runners.workloads.brokers import Broker
from quesadilla.runners.workloads.protocol import Executor, Workload
from quesadilla.runners.workloads.workers import Worker


class BaseRunnerLoggingExtras(RunnerLoggingExtras):
    task_queue: TaskQueueLoggingExtras


@dataclass(frozen=True)
class BaseRunner(Runner, metaclass=ABCMeta):
    task_queue: TaskQueue

    config: RunnerConfig = field(kw_only=True)

    id: ulid.ULID = field(default_factory=ulid.ULID, init=False)

    logger: logging.Logger = field(init=False, repr=False)
    stopped: Event = field(init=False, repr=False)

    workers: list[Executor[Worker]] = field(
        init=False, repr=False, default_factory=list
    )
    brokers: list[Executor[Broker]] = field(
        init=False, repr=False, default_factory=list
    )

    @cached_property
    def name(self) -> str:
        return (
            f"{str(self.task_queue.namespace)}.{str(self.task_queue.name)}."
            f"runner.{str(self.id)}"
        )

    @cached_property
    def logging_extras(self) -> BaseRunnerLoggingExtras:
        return {
            "id": str(self.id),
            "task_queue": self.task_queue.logging_extras,
        }

    def __post_init__(self) -> None:
        super().__setattr__("logger", logging.getLogger(self.name))
        super().__setattr__("stopped", self.make_event())

        self.workers.extend(
            self.make_workload(
                Worker(
                    runner=self,
                    task_queue=self.task_queue,
                    stopped=self.make_event(),
                    closed=self.make_event(),
                    scheduled_tasks=self.make_queue(),
                    results=self.make_queue(),
                )
            )
            for _ in range(self.config.workers)
        )
        self.brokers.extend(
            self.make_workload(
                Broker(
                    runner=self,
                    task_queue=self.task_queue,
                    stopped=self.make_event(),
                    closed=self.make_event(),
                    worker_selector=self.config.worker_selector_class(
                        [executor.workload for executor in self.workers]
                    ),
                )
            )
            for _ in range(self.config.brokers)
        )

    def wait(self) -> None:
        self.stopped.wait()

    def signal(self, sig: int | None = None) -> None:
        if not self.stopped.is_set():  # pragma: nobranch
            s = signal.Signals(sig) if sig is not None else sig
            if s is not None:  # pragma: nobranch
                self.logger.warning(
                    f"Received signal {s.name}", extra=self.logging_extras
                )
            if s in frozenset(
                (signal.SIGTERM, signal.SIGINT, None)
            ):  # pragma: nobranch
                self.stopped.set()

    def __enter__(self) -> Self:
        if self.stopped.is_set():
            raise RuntimeError("This runner has already been run")

        self.logger.info("Runner starting...", extra=self.logging_extras)
        for executor in chain(self.brokers, self.workers):
            executor.start()

        return self

    def __exit__(self, *_: Any, **__: Any) -> None:
        self.logger.info("Runner stopping...", extra=self.logging_extras)

        # the order in which we stop executors is very important
        # first, we close both workers and brokers to stop accepting new work
        for executor in chain(self.workers, self.brokers):
            executor.close()
        # then we stop and join workers to flush all scheduled work
        for executor in self.workers:
            executor.stop()
        for executor in self.workers:
            executor.join()
        # finally, we stop and join brokers to send results to the connector
        for executor in self.brokers:
            executor.stop()
        for executor in self.brokers:
            executor.join()

        self.logger.info("Runner exited gracefully", extra=self.logging_extras)
        self.stopped.set()

    @abstractmethod
    def make_workload[T: (Workload)](self, workload: T) -> Executor[T]: ...

    @abstractmethod
    def make_queue(self) -> BasicQueue[SerializableTask[Any]]: ...

    @abstractmethod
    def make_event(self) -> Event: ...



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import signal
import threading
from contextlib import ExitStack
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any, Self

import ulid

from quesadilla.runners.runners.protocol import Event, Runner, RunnerLoggingExtras


class CompositeRunnerLoggingExtras(RunnerLoggingExtras):
    runners: list[RunnerLoggingExtras]


@dataclass(frozen=True)
class CompositeRunner(Runner):
    runners: list[Runner] = field(kw_only=True)

    id: ulid.ULID = field(default_factory=ulid.ULID, init=False)

    exit_stack: ExitStack = field(default_factory=ExitStack, init=False, repr=False)
    stopped: Event = field(
        kw_only=True, default_factory=threading.Event, init=False, repr=False
    )

    @cached_property
    def name(self) -> str:
        return f"runner.composite.{str(self.id)}"

    @cached_property
    def logging_extras(self) -> CompositeRunnerLoggingExtras:
        return {
            "id": str(self.id),
            "runners": [runner.logging_extras for runner in self.runners],
        }

    @cached_property
    def logger(self) -> logging.Logger:
        return logging.getLogger(self.name)

    def wait(self) -> None:
        self.stopped.wait()

    def signal(self, sig: int | None = None) -> None:
        if not self.stopped.is_set():  # pragma: nobranch
            s = signal.Signals(sig) if sig is not None else sig
            if s in frozenset((signal.SIGTERM, signal.SIGINT)):  # pragma: nobranch
                for runner in self.runners:
                    runner.signal(sig)
                self.stopped.set()

    def __enter__(self) -> Self:
        if self.stopped.is_set():
            raise RuntimeError("This runner has already been run")

        self.logger.info("Runner starting...", extra=self.logging_extras)
        for runner in self.runners:
            self.exit_stack.enter_context(runner)

        return self

    def __exit__(self, *_: Any, **__: Any) -> None:
        self.logger.info("Runner stopping...", extra=self.logging_extras)
        self.exit_stack.close()
        self.logger.info("Runner exited gracefully", extra=self.logging_extras)
        self.stopped.set()



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property
from typing import Protocol, TypedDict

from quesadilla.core.logging import QueuedTaskLoggingExtras, TaskQueueLoggingExtras
from quesadilla.core.task_queues import GenericQueuedTask

from quesadilla.runners.runners.protocol import RunnerLoggingExtras


class Event(Protocol):
    def is_set(self) -> bool: ...  # pragma: nocover

    def set(self) -> None: ...  # pragma: nocover

    def wait(self, timeout: float | None = None) -> bool: ...  # pragma: nocover


class WorkloadLoggingExtras(TypedDict):
    task_queue: TaskQueueLoggingExtras
    id: str
    runner: RunnerLoggingExtras


class ExecutedQueuedTaskLoggingExtras(QueuedTaskLoggingExtras):
    workload: WorkloadLoggingExtras


class Workload:
    @cached_property
    def name(self) -> str: ...  # pragma: nocover

    @cached_property
    def logging_extras(self) -> WorkloadLoggingExtras: ...  # pragma: nocover

    def main_loop(self) -> None: ...  # pragma: nocover

    def close(self) -> None: ...  # pragma: nocover

    def stop(self) -> None: ...  # pragma: nocover

    def get_queued_task_logging_extras(
        self, queued_task: GenericQueuedTask
    ) -> ExecutedQueuedTaskLoggingExtras: ...  # pragma: nocover


class Executor[T: Workload](Protocol):
    workload: T

    def start(self) -> None: ...  # pragma: nocover

    def close(self) -> None: ...  # pragma: nocover

    def stop(self) -> None: ...  # pragma: nocover

    def join(self, timeout: float | None = None) -> None: ...  # pragma: nocover
